from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.timestream_query_result import QueryResult_TimestreamVariables
import json


class timeSeriesDb_queryDatabaseWithTimestreamQuery(Query):

    def __init__(self, client, debug, timeSeriesDbId, queryString, maxRows, nextToken, clientToken):
        super().__init__(client, debug, "timeSeriesDb_queryDatabaseWithTimestreamQuery",
                         QueryResult_TimestreamVariables)
        if timeSeriesDbId is None:
            raise MissingParameterException("Missing timeSeriesDbId")
        self.timeSeriesDbId = timeSeriesDbId
        self.queryString = json.dumps(queryString)
        self.maxRows = maxRows
        self.nextToken = nextToken
        self._clientToken = clientToken

    def _getQueryString(self):
        query = f'query {self.queryName} {{ {self.queryName}(timeSeriesDbId: "{self.timeSeriesDbId}", queryString: {self.queryString}'
        if self.maxRows:
            query += f', maxRows: {self.maxRows}'
        if self.nextToken:
            query += f', nextToken: "{self.nextToken}"'
        if self._clientToken:
            query += f', clientToken: "{self._clientToken}"'
        query += f') {self.resultType.getQueryString()} }}'
        return query

    def performQuery(self) -> str:
        print(
            f'Querying time series db {self.timeSeriesDbId}{f" with nextToken" if self.nextToken is not None else ""}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> QueryResult_TimestreamVariables:
        super()._processResult()
        try:
            self.result = QueryResult_TimestreamVariables(
                self._result['data'][self.queryName], self._client, self._debug, self.timeSeriesDbId, self.queryString, self.maxRows)
            print(f'{self.result}')
            return self.result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals


class Service:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionaryRoot(dict)

    def __str__(self):
        if self.serviceType == "DATA_STORE":
            return f"""{self.dataStore}"""
        elif self.serviceType == "TIME_SERIES_DB":
            return f"""{self.timeSeriesDb}"""
        else:
            return f"""{self.serviceType} - id: {self.serviceId}"""

    def buildFromDictionaryRoot(self, dict):
        from .data_store import DataStore
        from .time_series_db import TimeSeriesDb

        if dict is None:
            raise MissingDataInResultException

        if 'serviceType' in dict:
            self.serviceId = dict['serviceId']
            self.serviceType = dict['serviceType']
            if self.serviceType == "DATA_STORE":
                self.dataStore = DataStore(
                    dict['dataStore'], self._client, self._debug)
            elif self.serviceType == "TIME_SERIES_DB":
                self.timeSeriesDb = TimeSeriesDb(
                    dict['timeSeriesDb'], self._client, self._debug)
        elif 'dataStoreId' in dict:
            self.serviceId = dict['dataStoreId']
            self.serviceType = "DATA_STORE"
        elif 'timeSeriesDbId' in dict:
            self.serviceId = dict['timeSeriesDbId']
            self.serviceType = "DATA_STORE"
        else:
            raise ServiceTypeNotAvailable

    @staticmethod
    def getCorrectServiceTypeFromServiceObject(dict, client, debug):
        from .data_store import DataStore
        from .time_series_db import TimeSeriesDb
        if dict["serviceType"] == "DATA_STORE":
            return DataStore(dict['dataStore'], client, debug)
        elif dict["serviceType"] == "TIME_SERIES_DB":
            return TimeSeriesDb(dict['timeSeriesDb'], client, debug)
        else:
            raise ServiceTypeNotAvailable

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        from .data_store import DataStore
        from .time_series_db import TimeSeriesDb
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}serviceId
{tabStr}serviceType
{tabStr}{f"timeSeriesDb {TimeSeriesDb.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}{f"dataStore {DataStore.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals


class PresignedUrl:

    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.url = dict['url']
        self.uploadId = dict['uploadId']
        self.itemId = dict['itemId']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}url
{tabStr}uploadId
{tabStr}itemId
{tabStr[0:-4]}}} """




from ..exceptions import *
from .. import globals
from .endpoint_telemetry import EndpointTelemetry


class EndpointTelemetryDataRecord:

    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    # def __str__(self):
    #     result = f"""Endpoint last values:"""
    # # Record: iotEndpointId: {self.record.iotEndpointId} timestamp: {self.record.timestamp} ({self.record.timeUnit})
    # # Errors:"""
    # #     for error in self.errors:
    # #         result += f"""
    # #     - {error.message}"""
    #     return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException

        self.telemetryModel = EndpointTelemetry(
            dict['telemetryModel'], self._client, self._debug)
        self.timestamp = dict['timestamp']
        self.value = dict['value']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}{f"telemetryModel {EndpointTelemetry.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}timestamp
{tabStr}value
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals


class IotDataRecordError:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""IotDataRecordError:
        errorType: {self.errorType}
        errorMessage: {self.errorMessage}
"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.errorType = dict['errorType']
        self.errorMessage = dict['errorMessage']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}errorType
{tabStr}errorMessage
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_endpoints_last_values_query_result import IotEndpointsLastValuesQueryResult
import json


class iot_listEndpointLastValues(Query):

    def __init__(self, client, debug, iotEndpointId, clientToken=None, nextToken=None):
        super().__init__(client, debug, "iot_listEndpointLastValues",
                         IotEndpointsLastValuesQueryResult)
        if iotEndpointId is None:
            raise MissingParameterException("Missing iotEndpointId parameter")
        self.iotEndpointId = iotEndpointId
        self.nextToken = nextToken
        self.clientToken = clientToken

    def _getQueryString(self):
        query = f'query {self.queryName} {{ {self.queryName}(iotEndpointId: "{self.iotEndpointId}"'
        if self.nextToken:
            query += f', nextToken: "{self.nextToken}"'
        if self.clientToken:
            query += f', clientToken: "{self._clientToken}"'
        query += f') {self.resultType.getQueryString()} }}'
        return query

    def performQuery(self) -> str:
        print(
            f'Listing last values for telemetry of endpoint {self.iotEndpointId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> IotEndpointsLastValuesQueryResult:
        super()._processResult()
        try:
            self.result = IotEndpointsLastValuesQueryResult(
                self._result['data'][self.queryName], self._client, self._debug)
            print(f'{self.result}')
            return self.result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals
import requests
from ..data_types.user import User
from ..data_types.data_store_object import DataStoreObject
import os


class DataStoreFile(DataStoreObject):
    def __init__(self, dict, client, debug, directory=None):
        super().__init__(dict, client, debug)
        self.dataStoreFile = self
        self._client = client
        self._debug = debug
        self.directory = directory
        if dict is None:
            return
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""File: {self.dataStoreFileId}
  name: {self.getFilename()}
  description: {self.description}
  dataStoreId: {self.dataStoreId}
  directoryId: {self.dataStoreDirectoryId}
  size: {self.storageSizeBytes} bytes"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.dataStoreFileId = dict['dataStoreFileId']
        self.dataStoreId = dict['dataStoreId']
        self.name = dict['name']
        self.description = dict['description'] if 'description' in dict else ''
        self.ownerUserId = dict['ownerUserId']
        self.fileExtension = dict['fileExtension']
        self.storageSizeBytes = dict['storageSizeBytes']
        self.dataStoreDirectoryId = dict['dataStoreDirectoryId']
        self.s3ConfirmedUpload = dict['s3ConfirmedUpload']
        self.lastModified = dict['lastModified']
        self.contentType = dict['contentType']
        # self.dataViewIds = dict['dataViewIds']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}dataStoreFileId
{tabStr}dataStoreId
{tabStr}dataStoreDirectoryId
{tabStr}name
{tabStr}fileExtension
{tabStr}description
{tabStr}storageSizeBytes
{tabStr}ownerUserId
{tabStr}s3ConfirmedUpload
{tabStr}lastModified
{tabStr}contentType
{tabStr}dataViewIds
{tabStr[0:-4]}}} """

    def getFileUrl(self, dataStoreFileId):
        from ..graphql_queries.storage_getDataStoreFileDownloadUrl import storage_getDataStoreFileDownloadUrl
        if self._debug:
            print('DEBUG - Retrieving file download url...')
        query = storage_getDataStoreFileDownloadUrl(
            self._client, self._debug, self.dataStoreId, dataStoreFileId)
        result = query.performQuery()
        if self._debug:
            print('DEBUG - File download url received...')
        return result

    def getFilename(self):
        temp = self.name
        if self.fileExtension is not None:
            temp += '.' + self.fileExtension
        return temp

    def downloadAndWriteFile(self, url, writeLocation):
        response = requests.get(url)
        fullWritePath = self.createWritePath(writeLocation, self.getFilename())
        open(fullWritePath, "wb").write(response.content)
        print(f'Wrote file to {fullWritePath}')

    @staticmethod
    def createWritePath(writeLocation, cloudFilename):
        # if location is a directory, append the cloudFilename
        if os.path.isdir(writeLocation):
            return os.path.join(writeLocation, cloudFilename)
        # if not, use the writeLocation
        elif os.path.isfile(writeLocation):
            return writeLocation
        else:
            return writeLocation

    @staticmethod
    def uploadFile(url, filePath, debug):
        if debug:
            print(f'DEBUG - Uploading file {filePath}')
        try:
            result = requests.put(url, data=open(filePath, 'rb'))
            print(f'upload file returned status code: {result.status_code}')
            if result.ok:
                if debug:
                    print("DEBUG - File uploaded successfully")
            else:
                raise FileUploadException
        except Exception as err:
            print(f"Upload failed: {err}")
            raise FileUploadException(err)

    @staticmethod
    def getFilenameSegments(name):
        fileExtension = '.'.join(name.split('.')[1:])
        fileRoot = name.split('.')[0]
        return (fileRoot, fileExtension)

    @staticmethod
    def createAndUploadFile(client, debug, localFilePath, dataStoreId, dataStoreDirectoryId, name=None, description=None):
        from ..graphql_mutations.storage_createDataStoreFile import storage_createDataStoreFile

        # Check if file exists in local file system
        if not os.path.isfile(localFilePath):
            raise FileNotFoundException
        if not name:
            name = localFilePath.split('/')[-1]

        (fileRoot, fileExtension) = DataStoreFile.getFilenameSegments(name)

        # Create the file in the Digital Exchange and receive a presigned url to upload the file to
        mutation = storage_createDataStoreFile(
            client, debug, dataStoreId, dataStoreDirectoryId, fileRoot, fileExtension, description)
        presignedUrl = mutation.performMutation()
        dataStoreFileId = presignedUrl.itemId
        uploadResult = None
        try:
            if presignedUrl.url is None:
                raise Exception('Did not receive upload endpoint as expected')
            uploadResult = DataStoreFile.uploadFile(
                presignedUrl.url, localFilePath, debug)
        except FileUploadException as err:
            DataStoreFile.deleteDataStoreFile(
                client, debug, dataStoreFileId)
            raise err
        return DataStoreFile.getDataStoreFile(client, debug, dataStoreFileId)

    @staticmethod
    def deleteDataStoreFile(client, debug, dataStoreFileId):
        from ..graphql_mutations.storage_deleteDataStoreFile import storage_deleteDataStoreFile
        mutation = storage_deleteDataStoreFile(
            client, debug, dataStoreFileId)
        dataStoreFile = mutation.performMutation()
        return dataStoreFile

    @staticmethod
    def getDataStoreFile(client, debug, dataStoreFileId):
        from ..graphql_queries.storage_getDataStoreFile import storage_getDataStoreFile
        query = storage_getDataStoreFile(
            client, debug, dataStoreFileId)
        dataStoreFile = query.performQuery()
        return dataStoreFile

    @staticmethod
    def downloadDataStoreFile(client, debug, dataStoreFileId, writeLocation):
        from ..graphql_queries.storage_getDataStoreFile import storage_getDataStoreFile
        query = storage_getDataStoreFile(
            client, debug, dataStoreFileId)
        dataStoreFile = query.performQuery()
        presignedUrl = dataStoreFile.getFileUrl(dataStoreFileId)
        location = dataStoreFile.downloadAndWriteFile(
            presignedUrl.url, writeLocation)
        return dataStoreFile



from ..exceptions import *
from .. import globals
from .iot_attribute_value import IotAttributeValue


class IotDataRecord:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""DataRecord:
    iotEndpointId: {self.iotEndpointId}
"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.iotEndpointId = dict['iotEndpointId']
        self.timestamp = dict['timestamp']
        self.attributes = []
        for attribute in dict['attributes']:
            self.attributes.append(
                IotAttributeValue(attribute, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}iotEndpointId
{tabStr}timestamp
{tabStr}{f"attributes {IotAttributeValue.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """




# debug = True
# stage = "PRODUCTION"

# class GlobalVariables:
#     def __init__(self, debug, stage):
#         self.debug = debug
#         self.stage = stage

#         self.tab = "    "

#     def getTabStr(self, tabs=1):
#         tabStr = ""
#         for i in range(tabs):
#             tabStr += self.tab
#         return tabStr

tab = "    "


def getTabStr(tabs=1):
    tabStr = ""
    for i in range(tabs):
        tabStr += tab
    return tabStr



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.data_store import DataStore
from ..data_types.service import Service


class storage_listDataStores(Query):

    def __init__(self, client, debug):
        super().__init__(client, debug, "storage_listDataStores", DataStore)

    def _getQueryString(self):
        return super()._getQueryString()

    def performQuery(self) -> str:
        print('Listing your data stores...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.dataStores = []

            for currentDataStore in self._result['data'][self.queryName]:
                self.dataStores.append(
                    DataStore(currentDataStore, self._client, self._debug))
            i = 0

            for dataStore in self.dataStores:
                print(f'{i}: {dataStore}')
                i += 1
            return self.dataStores
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")




from .endpointUtil import *

from .graphql_mutations.iot_publish import iot_publish
from .graphql_mutations.iot_updateEndpointProperties import iot_updateEndpointProperties
from .graphql_mutations.iot_generateQueryResultExport import iot_generateQueryResultExport
from .graphql_mutations.iot_cancelQuery import iot_cancelQuery
from .graphql_queries.iot_getEndpoint import iot_getEndpoint
from .graphql_queries.iot_listEndpointLastValues import iot_listEndpointLastValues
from .graphql_queries.iot_queryByTimeRange import iot_queryByTimeRange
from .graphql_subscriptions.onNotifyIotQueryExportComplete import onNotifyIotQueryExportComplete
from .data_types.iot_data_record_input import IotDataRecordInput
from .data_types.iot_attribute_value_input import IotAttributeValueInput
from .data_types.endpoint_property_input import EndpointPropertyInput
from .data_types.endpoint_query_filter_input import EndpointQueryFilterInput
import time
import signal
import datetime


class IoTUtil(object):

    def __init__(self, debug, client):
        self._debug = debug
        self._client = client
        self.endpoint = EndpointUtil(debug, client)

    def publish(self, iotEndpointId, data):
        inputData = []
        for record in data:
            attributes = []
            if "timestamp" not in record:
                record["timestamp"] = f'{round(time.time()*1000)}'
            if "timeUnit" not in record:
                record["timeUnit"] = f'MILLISECONDS'
            IotDataRecordInput.checkDictForCorrectFields(record)
            for attributeName, value in (record["attributes"]).items():
                attributes.append(IotAttributeValueInput(
                    attributeName, value))
            inputData.append(IotDataRecordInput(
                record["timestamp"], record["timeUnit"] if "timeUnit" in record else "MILLISECONDS", attributes))

        mutation = iot_publish(
            self._client, self._debug, iotEndpointId, inputData)
        result = mutation.performMutation()

        resultDict = result.resultDict

        return {
            'recordsWritten': resultDict["recordsWritten"],
            'recordsFailed': resultDict["failedRecords"]
        }

    def updateEndpointProperties(self, iotEndpointId, properties):
        for key in properties:
            if 'value' not in properties[key]:
                raise Exception(f"Missing 'value' in properties['{key}']")

        propertiesInput = []
        for key in properties:
            propertiesInput.append(EndpointPropertyInput(
                key,
                properties[key]["value"],
                properties[key]["timestamp"] if 'timestamp' in properties[key] else f'{round(time.time()*1000)}'))

        mutation = iot_updateEndpointProperties(
            self._client, self._debug, iotEndpointId, propertiesInput)
        result = mutation.performMutation()
        return result

    def queryByTimeRange(self, iotEndpointId, attributeNames, startTime, endTime, resultWriteLocation=None):

        date = datetime.datetime.strptime(startTime, '%Y-%m-%dT%H:%M:%S.%fZ')
        startTime = int(
            (date - datetime.datetime(1970, 1, 1)).total_seconds()*1000)
        date = datetime.datetime.strptime(endTime, '%Y-%m-%dT%H:%M:%S.%fZ')
        endTime = int(
            (date - datetime.datetime(1970, 1, 1)).total_seconds()*1000)

        filter = EndpointQueryFilterInput(iotEndpointId, attributeNames)
        endpointFilterInput = [filter]
        query = iot_queryByTimeRange(
            self._client, self._debug, endpointFilterInput, startTime, endTime)
        result = query.performQuery()
        queryId = result.queryId
        print(
            f'received query result chunk {result.resultChunkIndex}')

        def quitHandler(signum, frame):
            # Stop query
            if queryId is not None:
                print("Cancelling query....")
                mutation = iot_cancelQuery(
                    self._client, self._debug, queryId)
                cancelResult = mutation.performMutation()
            exit(1)

        signal.signal(signal.SIGINT, quitHandler)

        while result.nextToken is not None:
            query = iot_queryByTimeRange(
                self._client, self._debug, endpointFilterInput, startTime, endTime, result.nextToken, result.clientToken, result.queryString, result.queryId)
            result = query.performQuery()
            print(
                f'received query result chunk {result.resultChunkIndex}')

        # get onNotifyObject
        subscription = self._getSubscriptionToQueryExportFileCompletion(
            queryId)

        # Generate export file
        mutationResult = self._generateQueryResultFile(queryId, "CSV")
        if self._debug:
            print(f'mutation result {mutationResult}')

        # wait for subscription notification
        exportFileResult = subscription.waitForResponse(240)
        if self._debug:
            print(exportFileResult)

        exportFileResult.downloadAndWriteFile(
            exportFileResult.url, resultWriteLocation)

        return True

    def getEndpoint(self, iotEndpointId):
        query = iot_getEndpoint(
            self._client, self._debug, iotEndpointId)
        result = query.performQuery()
        return result

    def listEndpointLastValues(self, iotEndpointId):
        query = iot_listEndpointLastValues(
            self._client, self._debug, iotEndpointId)
        result = query.performQuery()
        resultArray = []
        for entry in result.data[0].telemetryValues:
            resultArray.append({
                "attributeName": entry.telemetryModel.name,
                "lastValue": entry.value,
                "timestamp": entry.timestamp
            })
        return resultArray

    def _getSubscriptionToQueryExportFileCompletion(self, queryId):
        subscription = onNotifyIotQueryExportComplete(
            self._client, self._debug, queryId)
        subscribedThread = subscription.initiateSubscriptionOnNewThread()
        return subscription

    def _generateQueryResultFile(self, queryId, fileFormat="CSV"):
        mutation = iot_generateQueryResultExport(
            self._client, self._debug, queryId, fileFormat)
        result = mutation.performMutation()
        return result



from .organization import Organization
from ..exceptions import *
from .. import globals


class User:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.firstName = dict['firstName']
        self.lastName = dict['lastName']
        self.userId = dict['userId']
        self.email = dict['email']
        self.mobile = dict['mobile']
        self.userAccountState = dict['userAccountState']
        self.lastActive = dict['lastActive']
        self.userRoles = dict['userRoles']
        self.organization = Organization(
            dict['organization'], self._client, self._debug)

    def __str__(self):
        return f"""User
    name: {self.firstName + " " + self.lastName}
    userId: {self.userId}
    userRoles: {self.userRoles}
    organization name: {self.organization.name}
    organizationId: {self.organization.organizationId}
"""

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}firstName
{tabStr}lastName
{tabStr}userId
{tabStr}email
{tabStr}lastActive
{tabStr}mobile
{tabStr}userAccountState
{tabStr}userRoles
{tabStr}{f"organization {Organization.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.time_series_db import TimeSeriesDb


class timeSeriesDb_getDatabase(Query):

    def __init__(self, client, debug, timeSeriesDbId):
        super().__init__(client, debug, "timeSeriesDb_getDatabase", TimeSeriesDb)
        if timeSeriesDbId is None:
            raise MissingParameterException("Missing timeSeriesDbId")
        self.timeSeriesDbId = timeSeriesDbId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(timeSeriesDbId: "{self.timeSeriesDbId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(f'Retrieving time series db {self.timeSeriesDbId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> TimeSeriesDb:
        super()._processResult()
        try:
            self.database = TimeSeriesDb(
                self._result['data'][self.queryName], self._client, self._debug)
            print(f'{self.database}')
            return self.database
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_data_record_input import IotDataRecordInput
from ..data_types.publish_iot_data_result import PublishIotDataResult


class iot_publish(Mutation):

    def __init__(self, client, debug, iotEndpointId, data):
        super().__init__(client, debug, "iot_publish", PublishIotDataResult)
        self.iotEndpointId = iotEndpointId
        self.data = data

    def _getMutationString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        return f"""mutation {self.mutationName} {{ 
{tabStr}{self.mutationName}(
{tabStr}    iotEndpointId: "{self.iotEndpointId}", 
{tabStr}    data: [{','.join([record.getMutationParameterString(tabs+1) for record in self.data])}]) {self.resultType.getQueryString(tabs+1)} 
}}"""

    def performMutation(self):
        print(f"Publishing data to endpoint {self.iotEndpointId}...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.response = PublishIotDataResult(
                self._result['data'][self.mutationName], self._client, self._debug)
            print("Publish succeeded")
            return self.response
        except Exception as e:
            raise e



# subscription_client.py

from base64 import b64encode, decode
from datetime import datetime
# from lib2to3.pgen2 import token
from uuid import uuid4

import websocket
import threading

import json
import ssl

global globToken
global SUB_ID
global timeoutTimer
global timeout_interval
global graphqlSubscription
global HOST
global subscriptionOperationName


class Subscription:
    def __init__(self, debug, subscriptionName, subscriptionQueryString, API_URL, token, subscriptionMessageQueue, subscriptionErrorQueue, onMessage=None, onOpen=None, onClose=None, onError=None):
        """GraphQL subscription Registration object"""
        self._debug = debug
        self.subscriptionName = subscriptionName
        self.subscriptionQueryString = subscriptionQueryString
        self._API_URL = API_URL
        self._token = token
        self.subscriptionMessageQueue = subscriptionMessageQueue
        self.subscriptionErrorQueue = subscriptionErrorQueue

        self.graphqlSubscription = json.dumps({
            "query": self.subscriptionQueryString,
            "variables": {}
        })

        # Discovered values from the AppSync endpoint (API_URL)
        self.WSS_URL = self._API_URL.replace('https', 'wss').replace(
            'appsync-api', 'appsync-realtime-api')
        self.HOST = self._API_URL.replace(
            'https://', '').replace('/graphql', '')

        self.SUB_ID = str(uuid4())
        self.timeoutTimer = None
        self.timeout_interval = 10

        self.headers = {'HOST': self.HOST,
                        'Authorization': self._token}

        # Set up the connection URL, which includes the Authentication Header
        #   and a payload of '{}'.  All info is base 64 encoded
        self.connectionUrl = self.WSS_URL + '?header=' + \
            headerEncode(self.headers) + '&payload=e30='

        # self.performSubscribe(self)

        global globToken
        global SUB_ID
        global timeoutTimer
        global timeout_interval
        global graphqlSubscription
        global HOST
        global subscriptionOperationName

        globToken = self._token
        SUB_ID = self.SUB_ID
        timeoutTimer = self.timeoutTimer
        timeout_interval = self.timeout_interval
        graphqlSubscription = self.graphqlSubscription
        HOST = self.HOST
        subscriptionOperationName = self.subscriptionName

    def connect(self):
        # Create the websocket connection to AppSync's real-time endpoint
        #  also defines callback functions for websocket events
        #  NOTE: The connection requires a subprotocol 'graphql-ws'

        # Uncomment to see socket bytestreams
        # websocket.enableTrace(True)
        if self._debug:
            print('Connecting to: ' + self.connectionUrl)
        self.ws = websocket.WebSocketApp(self.connectionUrl,
                                         subprotocols=['graphql-ws'],
                                         on_open=self.onOpen,
                                         on_message=self.onMessage,
                                         on_error=self.onError,
                                         on_close=self.onClose)
        # ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})
        self.ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})

    def stop(self):
        self.ws.close()

    def onMessage(self, ws, message):
        """Socket Event Callbacks, used in WebSocketApp Constructor"""

        global globToken
        global SUB_ID
        global timeoutTimer
        global timeout_interval
        global graphqlSubscription
        global HOST
        global subscriptionOperationName

        if self._debug:
            print('### message ###')
            print(f'<<  {message}')

        message_object = json.loads(message)
        message_type = message_object['type']

        if (message_type == 'ka'):
            resetTimer(ws)

        elif (message_type == 'connection_ack'):
            timeout_interval = int(json.dumps(
                message_object['payload']['connectionTimeoutMs']))
            register = {
                'id': SUB_ID,
                'payload': {
                    'data': graphqlSubscription,
                    # 'query': graphqlSubscription.query,
                    # 'variables': graphqlSubscription.variables,
                    'operationName': subscriptionOperationName,
                    'extensions': {
                        'authorization': {
                            'host': HOST,
                            'Authorization': globToken,
                            'Content-Type': 'application/graphql'
                        }
                    }
                },
                'type': 'start'
            }
            start_sub = json.dumps(register)
            if self._debug:
                print(f'>> {start_sub}')
            ws.send(start_sub)
        # elif (message_type == 'start_ack'):
        #     print
        elif (message_type == 'data'):
            deregister = {
                'type': 'stop',
                'id': SUB_ID
            }
            end_sub = json.dumps(deregister)
            if self._debug:
                print(f'>> {end_sub}')
            ws.send(end_sub)
            self.subscriptionMessageQueue.put(message_object)

        elif (message_object['type'] == 'error'):
            print(f"Error from AppSync: {message_object['payload']}")
            self.subscriptionErrorQueue.put(message_object)

    def onError(self, ws, error):
        if self._debug:
            print('### subscription websocket error ###')
            print(error)
            print()
        self.subscriptionErrorQueue.put(error)

    def onClose(self, ws, sec=None, thir=None, forth=None):
        if self._debug:
            print('### subscription websocket closed ###')
            print()

    def onOpen(self, ws):
        init = {
            'type': 'connection_init'
        }
        init_conn = json.dumps(init)
        if self._debug:
            print('### subscription websocket opened ###')
            print(f'>> {init_conn}')
            print()
        ws.send(init_conn)


def resetTimer(ws):
    """reset the keep alive timeout daemon thread"""
    global globToken
    global SUB_ID
    global timeoutTimer
    global timeout_interval
    global graphqlSubscription
    global HOST

    if (timeoutTimer):
        timeoutTimer.cancel()
    timeoutTimer = threading.Timer(
        timeout_interval, lambda: ws.close())
    timeoutTimer.daemon = True
    timeoutTimer.start()


# Calculate UTC time in ISO format (AWS Friendly): YYYY-MM-DDTHH:mm:ssZ


def headerTime():
    return datetime.utcnow().isoformat(sep='T', timespec='seconds') + 'Z'

# Encode Using Base 64


def headerEncode(header_obj):
    return b64encode(json.dumps(header_obj).encode('utf-8')).decode('utf-8')



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.user import User


class user_getUserSession(Query):

    def __init__(self, client, debug):
        super().__init__(client, debug, "user_getUserSession", User)

    def _getQueryString(self):
        tabs = "    "
        return f"""
query {self.queryName} {{ {self.queryName} {{ 
{tabs}user {self.resultType.getQueryString(2)} 
}}}}"""

    def performQuery(self):
        print('Getting your user information...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.user = User(
                self._result['data'][self.queryName]['user'], self._client, self._debug)
            print(self.user)
            return self.user
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")




from ..exceptions import *
from .. import globals
from .endpoint_telemetry_data_record import EndpointTelemetryDataRecord


class EndpointLastValues:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    # def __str__(self):
    #     result = f"""Endpoint last values:"""
    # # Record: iotEndpointId: {self.record.iotEndpointId} timestamp: {self.record.timestamp} ({self.record.timeUnit})
    # # Errors:"""
    # #     for error in self.errors:
    # #         result += f"""
    # #     - {error.message}"""
    #     return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException

        self.iotEndpointId = dict["iotEndpointId"]
        self.telemetryValues = []
        for entry in dict["telemetryValues"]:
            self.telemetryValues.append(EndpointTelemetryDataRecord(
                entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}iotEndpointId
{tabStr}{f"telemetryValues {EndpointTelemetryDataRecord.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from .user import User


class CollaborativeMemberUser:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.permission = dict['permission']
        self.user = User(dict['user'], self._client, self._debug)

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}permission
{tabStr}{f"user {User.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.data_store_file import DataStoreFile


class storage_getDataStoreFile(Query):

    def __init__(self, client, debug, dataStoreFileId):
        super().__init__(client, debug, "storage_getDataStoreFile", DataStoreFile)
        if dataStoreFileId is None:
            raise MissingParameterException("Missing dataStoreFileId")
        self.dataStoreFileId = dataStoreFileId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(dataStoreFileId: "{self.dataStoreFileId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        if self._debug:
            print(f"DEBUG - Retrieving file {self.dataStoreFileId}...")
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> DataStoreFile:
        super()._processResult()
        try:
            result = DataStoreFile(
                self._result['data'][self.queryName], self._client, self._debug)
            return result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from .data_types.service import Service
from .data_types.data_store_directory import DataStoreDirectory
from .data_types.data_store_file import DataStoreFile
from .data_types.data_store import DataStore
from .graphql_mutations.storage_deleteDataStoreFile import storage_deleteDataStoreFile
from .graphql_mutations.storage_createDataStoreFile import storage_createDataStoreFile
from .graphql_queries.storage_getDataStoreDirectory import storage_getDataStoreDirectory
from .graphql_queries.storage_getDataStoreFile import storage_getDataStoreFile
from .graphql_queries.storage_getDataStore import storage_getDataStore
from .graphql_queries.storage_listDataStores import storage_listDataStores

import time
import signal


class DataStoreUtil(object):

    def __init__(self, debug, client):
        self._debug = debug
        self._client = client

    def listDataStores(self):
        """Lists the Data Stores belonging to the user"""
        query = storage_listDataStores(self._client, self._debug)
        result = query.performQuery()
        return result

    def getDataStore(self, dataStoreId) -> DataStore:
        """Retrieves the Data Store's metadata by dataStoreId"""
        query = storage_getDataStore(self._client, self._debug, dataStoreId)
        result = query.performQuery()
        return result

    def getDataStoreDirectory(self, dataStoreDirectoryId) -> DataStoreDirectory:
        """Retrieves a Data Store directory and it's contents by dataStoreDirectoryId"""
        return DataStoreDirectory.getDirectory(self._client, self._debug, dataStoreDirectoryId)

    def getDataStoreFileMeta(self, dataStoreFileId) -> DataStoreFile:
        """Retrieves a Data Store file's metadata by dataStoreFileId"""
        return DataStoreFile.getDataStoreFile(self._client, self._debug, dataStoreFileId)

    def downloadDataStoreFile(self, dataStoreFileId, writeLocation) -> DataStoreFile:
        """Retrieves a Data Store file's metadata and downloads and writes the file to the local file system"""
        return DataStoreFile.downloadDataStoreFile(self._client, self._debug, dataStoreFileId, writeLocation)

    def uploadFileToDataStore(self, dataStoreId, dataStoreDirectoryId, localFilePath, name=None, description=None) -> DataStoreFile:
        """Copies a local file to the Data Store and is placed in the directory matching dataStoreDirectoryId"""

        return DataStoreFile.createAndUploadFile(self._client,
                                                 self._debug, localFilePath, dataStoreId, dataStoreDirectoryId, name, description)

    def deleteDataStoreFile(self, dataStoreFileId) -> DataStoreFile:
        """Deletes the file from the Data Store"""
        return DataStoreFile.deleteDataStoreFile(self._client, self._debug, dataStoreFileId)



from ..exceptions import *
from .. import globals


class IotAttributeValueInput:
    def __init__(self, name, value):
        self.name = name
        self.value = value
        return

    def getMutationParameterString(self, tabs):
        tabStr = globals.getTabStr(tabs)
        result = f""" {{
{tabStr}name: "{self.name}",
{tabStr}value: "{self.value}"
{tabStr[0:-4]}}} """
        return result



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_endpoint import IotEndpoint


class iot_getEndpoint(Query):

    def __init__(self, client, debug, iotEndpointId):
        super().__init__(client, debug, "iot_getEndpoint", IotEndpoint)
        if iotEndpointId is None:
            raise MissingParameterException("Missing iotEndpointId")
        self.iotEndpointId = iotEndpointId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(iotEndpointId: "{self.iotEndpointId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(f'Retrieving iot endpoint {self.iotEndpointId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> IotEndpoint:
        super()._processResult()
        try:
            self.endpoint = IotEndpoint(
                self._result['data'][self.queryName], self._client, self._debug)
            return self.endpoint
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals
import json


class TimeSeriesQueryResult:
    def __init__(self, dict, client, debug, timeSeriesDbId, maxRows):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)
        self.timeSeriesDbId = timeSeriesDbId
        self.maxRows = maxRows

    def __str__(self):
        return f"""Query Result:
    clientToken: {self.clientToken}
    nextToken: {self.nextToken}"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException

        self.queryProgressPercentage = dict["queryProgressPercentage"]
        self.clientToken = dict["clientToken"]
        self.nextToken = dict["nextToken"]
        self.queryString = dict["queryString"]
        self.queryId = dict["queryId"]
        self.presignedUrl = dict["presignedUrl"]
        self.resultChunkIndex = dict["resultChunkIndex"]

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}queryProgressPercentage
{tabStr}clientToken
{tabStr}nextToken
{tabStr}queryString
{tabStr}queryId
{tabStr}presignedUrl
{tabStr}resultChunkIndex
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.time_series_db import TimeSeriesDb
from ..data_types.service import Service


class timeSeriesDb_listDatabases(Query):

    def __init__(self, client, debug):
        super().__init__(client, debug, "timeSeriesDb_listDatabases", TimeSeriesDb)

    def _getQueryString(self):
        return super()._getQueryString()

    def performQuery(self) -> str:
        print('Listing your time series databases...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.databases = []
            for currentDatabase in self._result['data'][self.queryName]:
                self.databases.append(TimeSeriesDb(
                    currentDatabase, self._client, self._debug))
            i = 0
            for database in self.databases:
                print(f'{i}: {database}')
                i += 1
            return self.databases
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.data_store_file import DataStoreFile


class storage_deleteDataStoreFile(Mutation):

    def __init__(self, client, debug, dataStoreFileId):
        super().__init__(client, debug, "storage_deleteDataStoreFile", DataStoreFile)
        self.dataStoreFileId = dataStoreFileId

    def _getMutationString(self):
        return f"""mutation {self.mutationName} {{ {self.mutationName}(dataStoreFileId: "{self.dataStoreFileId}") {self.resultType.getQueryString()} }}"""

    def performMutation(self):
        print(
            f"Deleting file {self.dataStoreFileId}...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.result = DataStoreFile(
                self._result['data'][self.mutationName], self._client, self._debug)
            print(f"File deleted")
            return self.result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of mutation {self.mutationName}")



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.server_response import ServerResponse


class timeSeriesDb_cancelDatabaseQuery(Mutation):

    def __init__(self, client, debug, timeSeriesDbId, queryId):
        super().__init__(client, debug, "timeSeriesDb_cancelDatabaseQuery", ServerResponse)
        self.timeSeriesDbId = timeSeriesDbId
        self.queryId = queryId

    def _getMutationString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        return f"""mutation {self.mutationName} {{ 
{tabStr}{self.mutationName}(
{tabStr}    timeSeriesDbId: "{self.timeSeriesDbId}", 
{tabStr}    queryId: "{self.queryId}") {self.resultType.getQueryString(tabs+1)} 
}}"""

    def performMutation(self):
        print("Cancelling query...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.response = ServerResponse(
                self._result['data'][self.mutationName], self._client, self._debug)
            return self.response
        except Exception as e:
            raise e




from .time_series_asset_data import TimeSeriesAssetData

from ..exceptions import *
from .. import globals
import json


class TimeSeriesAssetsAndLastValuesResult:
    def __init__(self, dict, client, debug, timeSeriesDbId):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)
        self.timeSeriesDbId = timeSeriesDbId

    def __str__(self):
        return f"""Query Progress Percentage: {self.queryProgressPercentage}
clientToken: {self.clientToken}
nextToken: {self.nextToken}
queryId: {self.queryId}
assets: 
{self.getDataString()}
"""

    def getDataString(self):
        result = ""
        for entry in self.data:
            result += f'{entry}'
        return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.queryProgressPercentage = dict["queryProgressPercentage"]
        self.clientToken = dict["clientToken"]
        self.nextToken = dict["nextToken"] if "nextToken" in dict else None
        self.queryId = dict["queryId"]
        self.resultChunkIndex = dict["resultChunkIndex"]
        self.data = []
        for i in range(0, len(dict["data"])):
            self.data.append(TimeSeriesAssetData(dict["data"][i],
                                                 self._client,
                                                 self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}queryProgressPercentage
{tabStr}clientToken
{tabStr}nextToken
{tabStr}queryId
{tabStr}resultChunkIndex
{tabStr}resultChunkIndex
{tabStr}{f"data {TimeSeriesAssetData.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals


class EndpointQueryFilterInput:
    def __init__(self, iotEndpointId, attributeNames):
        self.iotEndpointId = iotEndpointId
        self.attributeNames = attributeNames
        return

    def getMutationParameterString(self, tabs):
        tabStr = globals.getTabStr(tabs)
        result = f""" {{
{tabStr}iotEndpointId: "{self.iotEndpointId}",
{tabStr}attributeNames: ["""
        for index, name in enumerate(self.attributeNames):
            result += f'''"{name}"'''
            if index < len(self.attributeNames) - 1:
                result += ', '
        result += f"""]
{tabStr[0:-4]}}} """
        return result



from ..exceptions import *
from .. import globals
from .iot_data_record import IotDataRecord
from .iot_data_record_error import IotDataRecordError


class IotEndpointData:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        result = f"""iot data:
    iotEndpointId: {self.record.iotEndpointId} attributes: [{",".join(self.includedAttributeNames)}]"""
        for record in self.records:
            result += f"""
    {record}"""
        return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.iotEndpointId = dict['iotEndpointId']
        self.includedAttributeNames = dict['includedAttributeNames']
        self.records = []
        for entry in dict['records']:
            self.records.append(
                IotDataRecord(entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}iotEndpointId
{tabStr}includedAttributeNames
{tabStr}{f"records {IotDataRecord.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from .collaborative_member_user import CollaborativeMemberUser
from .user import User
from .organization import Organization


class CollaborativeMemberOrganization:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.organization = Organization(
            dict['organization'], self._client, self._debug)
        self.collaborativeId = dict['collaborativeId']
        self.memberState = dict['memberState']
        self.usersInCollaborative = []
        for user in dict['usersInCollaborative']:
            self.usersInCollaborative.append(
                CollaborativeMemberUser(user, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}collaborativeId
{tabStr}{f"organization {Organization.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}memberState
{tabStr}accessApprovedByUserId
{tabStr}{f"submittedByUser {User.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}submittedTimestamp
{tabStr}{f"usersInCollaborative {CollaborativeMemberUser.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.presigned_url import PresignedUrl


class storage_createDataStoreFile(Mutation):

    def __init__(self, client, debug, dataStoreId, dataStoreDirectoryId, name, fileExtension, description=None):
        super().__init__(client, debug, "storage_createDataStoreFile", PresignedUrl)
        self.dataStoreDirectoryId = dataStoreDirectoryId
        self.dataStoreId = dataStoreId
        self.name = name
        self.fileExtension = fileExtension
        self.description = description

    def _getMutationString(self):
        return f"""mutation {self.mutationName} {{ {self.mutationName}(dataStoreDirectoryId: "{self.dataStoreDirectoryId}", dataStoreId: "{self.dataStoreId}", name: "{self.name}", fileExtension: "{self.fileExtension}" {f', description: "{self.description}"' if self.description is not None else ''}) {self.resultType.getQueryString()} }}"""

    def performMutation(self):
        print("Creating data store file...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.presignedUrl = PresignedUrl(
                self._result['data'][self.mutationName], self._client, self._debug)
            print("Received upload endpoint for file...")
            return self.presignedUrl
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of mutation {self.mutationName}")



from ..exceptions import *
from .. import globals


class IotAttributeValue:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""attribute value:
    name: {self.name}
    value: {self.value}
"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.name = dict['name'] if 'name' in dict else None
        self.schemaType = dict['schemaType'] if 'schemaType' in dict else None
        self.value = dict['value'] if 'value' in dict else None

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}name
{tabStr}value
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from ..data_types.time_series_input_record import TimeSeriesInputRecord


class TimeSeriesPublishInput:
    def __init__(self, records):
        self.records = []
        if records:
            self.addRecords(records)
        return

    def addRecords(self, records):
        # Check that records are of the correct type
        for record in records:
            if not isinstance(record, TimeSeriesInputRecord):
                raise InvalidParameterException(
                    "Records array contains elements that are not of type TimeSeriesInputRecord")
        self.records.extend(records)

    def getMutationParameterString(self, tabs):
        tabStr = globals.getTabStr(tabs)
        result = f""" {{
{tabStr}records: [{','.join([record.getMutationParameterString(tabs+1) for record in self.records])}]
{tabStr[0:-4]}}} """
        return result



from time import sleep
from .subscription import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_query_export import IotQueryExport

from threading import Thread
from queue import Queue


class onNotifyIotQueryExportComplete(Subscription):

    def __init__(self, client, debug, queryId):
        super().__init__(client, debug,
                         "onNotifyIotQueryExportComplete", IotQueryExport)
        if queryId is None:
            raise MissingParameterException("Missing queryId")
        self.queryId = queryId

        self.subscriptionMessageQueue = Queue()
        self.subscriptionErrorQueue = Queue()

    def _getSubscriptionQueryString(self):
        return f'''subscription {self.subscriptionName} 
        {{ 
            {self.subscriptionName}(queryId: "{self.queryId}") {self.resultType.getQueryString()} 
        }}'''

    def initiateSubscriptionOnNewThread(self):
        print(f'Creating subscription for result export completion...')
        self.subscriptionClient = self._client.getSubscriptionClient(
            self.subscriptionName,
            self._getSubscriptionQueryString(),
            self.subscriptionMessageQueue,
            self.subscriptionErrorQueue)

        self.subscribedThread = Thread(
            target=self.subscriptionClient.connect)
        self.subscribedThread.start()

        return self.subscribedThread

    def endSubscription(self):
        self.subscriptionClient.stop()
        self.subscribedThread.join()

    def waitForResponse(self, maxWaitSeconds):
        i = 0
        done = False
        while not done:
            i += 1
            if self._debug:
                print('subscription message queue:')
                print(list(self.subscriptionMessageQueue.queue))
                print()

                print('subscription error queue')
                print(list(self.subscriptionErrorQueue.queue))
                print()

            if self.subscriptionMessageQueue.qsize() > 0:
                self.endSubscription()
                item = self.subscriptionMessageQueue.get()
                if self.queryId == item['payload']['data'][self.subscriptionName]['queryId']:
                    self._result = item['payload']
                    return self._processResult()
                else:
                    raise Exception(
                        "Missing data in export success notification")
            elif self.subscriptionErrorQueue.qsize() > 0:
                raise Exception(
                    f'Encountered error when waiting for query export file: {self.subscriptionErrorQueue.get()}')

            if i == maxWaitSeconds:
                self.endSubscription()
                raise Exception(
                    'Timed out waiting for export complete notification')
            else:
                print('waiting...')
                sleep(1)

    def _processResult(self) -> IotQueryExport:
        super()._processResult()
        try:
            print('received export complete notification')
            self.queryExport = IotQueryExport(
                self._result['data'][self.subscriptionName], self._client, self._debug)
            if self._debug:
                print(f'DEBUG - Result {self.queryExport}')
            return self.queryExport
        except Exception as e:
            raise e



from ..exceptions import *
from .. import globals
from .iot_data_record import IotDataRecord
from .iot_data_record_error import IotDataRecordError


class IotDataRecordWithErrors:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        result = f"""Record errors:
    Record: iotEndpointId: {self.record.iotEndpointId} timestamp: {self.record.timestamp} ({self.record.timeUnit})
    Errors:"""
        for error in self.errors:
            result += f"""
        - {error.message}"""
        return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.record = IotDataRecord(dict['record'], self._client, self._debug)
        self.errors = []
        for error in dict['errors']:
            self.errors.append(
                IotDataRecordError(error, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}{f"record {IotDataRecord.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}{f"errors {IotDataRecordError.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from .time_series_asset_attribute import TimeSeriesAssetAttribute


class TimeSeriesAssetData:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""asset: {self.name if self.name is not None else '<unknown>'}
    assetId: {self.assetId}
    attributes: {self.getAttributesString()}
"""

    def getAttributesString(self):
        result = ""
        for entry in self.attributes:
            result += f'{entry}'
        return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.assetId = dict['assetId']
        self.name = dict['name']
        self.attributes = []
        for attribute in dict['attributes']:
            self.attributes.append(
                TimeSeriesAssetAttribute(attribute, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}assetId
{tabStr}name
{tabStr}{f"attributes {TimeSeriesAssetAttribute.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_endpoint import IotEndpoint
from ..data_types.endpoint_property_input import EndpointPropertyInput


class iot_updateEndpointProperties(Mutation):

    def __init__(self, client, debug, iotEndpointId, properties):
        super().__init__(client, debug, "iot_updateEndpointProperties", IotEndpoint)
        self.iotEndpointId = iotEndpointId
        self.properties = properties

    def _getMutationString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        return f"""mutation {self.mutationName} {{ 
{tabStr}{self.mutationName}(
{tabStr}    iotEndpointId: "{self.iotEndpointId}", 
{tabStr}    properties: [{','.join([property.getMutationParameterString(tabs+1) for property in self.properties])}]) {self.resultType.getQueryString(tabs+1)} 
}}"""

    def performMutation(self):
        print(f"Updating properties for endpoint {self.iotEndpointId}...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.response = IotEndpoint(
                self._result['data'][self.mutationName], self._client, self._debug)
            return self.response
        except Exception as e:
            raise e



from ..exceptions import *
from .. import globals


class Mutation:
    def __init__(self, client, debug, mutationName, resultType):
        self._client = client
        self._debug = debug
        self._result = None
        self.mutationName = mutationName
        self.resultType = resultType
        return

    def _getMutationString(self):
        return f"mutation {self.mutationName} {{ {self.mutationName}() {{ {self.resultType.getQueryString(None, 4)} }} }}"

    def _processResult(self):
        if self._debug:
            print(f'DEBUG - processing mutation result: {self.mutationName}')
            print(f'DEBUG - Result object: {self._result}')
        if 'errors' in self._result and len(self._result['errors']) > 0:
            print('ERROR - Mutation encountered error.')
            print(f'ERROR - Mutation Response/Error: {self._result}')
            if 'errorType' in self._result['errors'][0] and ['errorType'] == "UnauthorizedException":
                raise UnauthorizedException
            elif 'message' in self._result['errors'][0]['message'] and self._result['errors'][0]['message'] == "Unauthorized":
                raise Unauthorized
            else:
                raise MutationFailed(
                    f"Mutation {self.mutationName} failed.", self._result)

        if self._result['data'][self.mutationName] is None:
            raise MissingDataInResultException



from ..exceptions import *
from .. import globals
from .time_series_asset_attribute_data import TimeSeriesAssetAttributeData


class TimeSeriesAssetAttribute:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""
        {self.name}
            attributeType: {self.attributeType}
            lastValue: {self.lastValue}
            lastValueTime: {self.lastValueTime}"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.attributeType = dict['attributeType']
        self.lastValue = dict['lastValue']
        self.lastValueTime = dict['lastValueTime']
        self.name = dict['name']
        self.data = []
        for entry in dict['data']:
            self.data.append(TimeSeriesAssetAttributeData(
                entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}attributeType
{tabStr}lastValue
{tabStr}lastValueTime
{tabStr}name
{tabStr}{f"data {TimeSeriesAssetAttributeData.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.time_series_query_result import TimeSeriesQueryResult
import json


class timeSeriesDb_queryDatabase(Query):

    def __init__(self, client, debug, timeSeriesDbId, assetsArray, startTime, endTime, maxRows=None, nextToken=None, clientToken=None, queryString=None):
        super().__init__(client, debug, "timeSeriesDb_queryDatabase",
                         TimeSeriesQueryResult)

        self.timeSeriesDbId = timeSeriesDbId
        self.assetsArray = assetsArray
        self.maxRows = maxRows

        if not isValidTimestampFormat(startTime):
            raise InvalidTimestampParameter("startTime", startTime)
        self.startTime = startTime
        if not isValidTimestampFormat(endTime):
            raise InvalidTimestampParameter("endTime", endTime)
        self.endTime = endTime

        self.queryString = queryString
        self.nextToken = nextToken
        self.clientToken = clientToken

        if self.nextToken is not None:
            self.queryType = "NEXT_TOKEN"
        else:
            self.queryType = "TIME_PERIOD"

    def getAttributeNamesListString(self, attributeNamesArray):

        result = f'"{attributeNamesArray[0]}"'
        for i in range(1, len(attributeNamesArray)):
            result += f', "{attributeNamesArray[i]}"'
        return result

    def _getQueryString(self):
        assetsArrayString = "["
        i = 0
        for entry in self.assetsArray:
            assetsArrayString += f'{{assetId: "{entry["assetId"]}", attributeNamesFilter: [{self.getAttributeNamesListString(entry["attributeNames"])}]}}'
            i += 1
            if i >= len(self.assetsArray):
                assetsArrayString += ','

        assetsArrayString += "]"

        query = f'''query {self.queryName} {{ {self.queryName}
            (timeSeriesDbId: "{self.timeSeriesDbId}", 
             queryType: TIME_PERIOD, 
             startTime: "{self.startTime}", 
             endTime: "{self.endTime}", 
             assets: {assetsArrayString}'''
        if self.maxRows:
            query += f', maxRows: {self.maxRows}'
        if self.nextToken:
            query += f', nextToken: "{self.nextToken}"'
        if self.queryString:
            query += f', queryString: {json.dumps(self.queryString)}'
        if self.clientToken:
            query += f', clientToken: "{self.clientToken}"'
        query += f') {self.resultType.getQueryString()} }}'
        return query

    def performQuery(self) -> str:
        print(f'Querying database {self.timeSeriesDbId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> TimeSeriesQueryResult:
        super()._processResult()
        try:
            self.result = TimeSeriesQueryResult(
                self._result['data'][self.queryName], self._client, self._debug, self.timeSeriesDbId, self.maxRows)
            return self.result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")


def isValidTimestampFormat(timestamp):
    return True



from ..exceptions import *
from .. import globals
import json


class QueryResult_TimestreamVariables:
    def __init__(self, dict, client, debug, timeSeriesDbId, queryString, maxRows):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)
        self.timeSeriesDbId = timeSeriesDbId
        self.queryString = queryString
        self.maxRows = maxRows

    def __str__(self):
        return f"""Query Result
    clientToken: {self.clientToken}
    nextToken: {self.nextToken}
    columnInfo: {len(self.columnInfo)} columns
    dataRows: {len(self.dataRows)} data rows
"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.timestreamResultsJSONString = dict["resultJSONString"]
        self.clientToken = dict["clientToken"]
        self.timestreamResults = json.loads(self.timestreamResultsJSONString)
        self.queryId = self.timestreamResults["QueryId"]
        self.queryStatus = self.timestreamResults["QueryStatus"]
        self.rows = self.timestreamResults["Rows"]
        self.dataRows = self.rows
        self.columnInfo = self.timestreamResults["ColumnInfo"]
        self.nextToken = self.timestreamResults["NextToken"] if "NextToken" in self.timestreamResults else None

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}resultJSONString
{tabStr}clientToken
{tabStr[0:-4]}}} """




from ..exceptions import *
from .. import globals
from .endpoint_last_values import EndpointLastValues
from .iot_endpoint_data import IotEndpointData


class IotQueryByTimeRangeResult:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    # def __str__(self):
    #     result = f"""Endpoint last values:"""
    # # Record: iotEndpointId: {self.record.iotEndpointId} timestamp: {self.record.timestamp} ({self.record.timeUnit})
    # # Errors:"""
    # #     for error in self.errors:
    # #         result += f"""
    # #     - {error.message}"""
    #     return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        if self._debug:
            print(dict)
        self.clientToken = dict['clientToken']
        self.nextToken = dict['nextToken'] if 'nextToken' in dict else None
        self.queryId = dict['queryId']
        self.resultChunkIndex = dict['resultChunkIndex']
        self.presignedUrl = dict['presignedUrl']
        self.queryString = dict['queryString']
        # self.data = []
        # if "data" in dict and dict["data"] != None:
        #     for entry in dict["data"]:
        #         self.data.append(IotEndpointData(
        #             entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}clientToken
{tabStr}nextToken
{tabStr}queryId
{tabStr}resultChunkIndex
{tabStr}presignedUrl
{tabStr}queryString
{tabStr[0:-4]}}} """
# {tabStr}{f"data {IotEndpointData.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
# {tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from .service import Service


class TimeSeriesDb(Service):
    def __init__(self, dict, client, debug):
        super().__init__(dict, client, debug)
        self.timeSeriesDb = self
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""name: {self.name}
    timeSeriesDbId: {self.timeSeriesDbId}
    description: {self.description}
    databaseName: {self.databaseName}
    tableName: {self.tableName}"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.timeSeriesDbId = dict['timeSeriesDbId']
        self.name = dict['name']
        self.description = dict['description']
        self.ownerUserId = dict['ownerUserId']
        self.databaseName = dict['databaseName']
        self.tableName = dict['tableName']

        self.serviceType = "TIME_SERIES_DB"
        self.serviceId = self.timeSeriesDbId

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}timeSeriesDbId
{tabStr}name
{tabStr}description
{tabStr}ownerUserId
{tabStr}databaseName
{tabStr}tableName
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals


class ServerResponse:

    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.httpStatusCode = dict["httpStatusCode"]
        self.errorMessage = dict["errorMessage"]

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}httpStatusCode
{tabStr}errorMessage
{tabStr[0:-4]}}} """




from ..exceptions import *
from .. import globals
from .endpoint_last_values import EndpointLastValues
from .endpoint_telemetry_data_record import EndpointTelemetryDataRecord


class IotEndpointsLastValuesQueryResult:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    # def __str__(self):
    #     #     result = f"""Endpoint last values:"""
    #     # # Record: iotEndpointId: {self.record.iotEndpointId} timestamp: {self.record.timestamp} ({self.record.timeUnit})
    #     # # Errors:"""
    #     # #     for error in self.errors:
    #     # #         result += f"""
    #     # #     - {error.message}"""
    #     #     return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException

        self.clientToken = dict['clientToken']
        self.nextToken = dict['nextToken']
        self.queryId = dict['queryId']
        self.resultChunkIndex = dict['resultChunkIndex']

        self.resultStoredInS3 = dict['resultStoredInS3']
        self.presignedUrl = dict['presignedUrl']
        self.data = []
        for entry in dict["data"]:
            self.data.append(EndpointLastValues(
                entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{ 
{tabStr}clientToken
{tabStr}nextToken
{tabStr}queryId
{tabStr}resultChunkIndex
{tabStr}resultStoredInS3
{tabStr}presignedUrl
{tabStr}{f"data {EndpointLastValues.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals
from .service import Service
from .collaborative_member_organization import CollaborativeMemberOrganization


class Collaborative:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""{self.name}
    collaborativeId: {self.collaborativeId}
    description: {self.description}
    {self._printMembers()}"""

    def _printMembers(self):
        if self.memberOrganizations is not None and len(self.memberOrganizations) > 0:
            result = "members: \n"
        else:
            return ""
        for org in self.memberOrganizations:
            if org.memberState == "APPROVED" or org.memberState == "HOST_ORGANIZATION":
                result += f"    - {org.organization.name} | {len(org.usersInCollaborative)} user{'s' if len(org.usersInCollaborative) > 1 else ''}\n"
        return result

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.collaborativeId = dict['collaborativeId']
        self.name = dict['name']
        self.description = dict['description']
        self.hostOrganizationId = dict['hostOrganizationId']
        self.memberOrganizations = []
        for org in dict['memberOrganizations']:
            self.memberOrganizations.append(
                CollaborativeMemberOrganization(org, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}collaborativeId
{tabStr}hostOrganizationId
{tabStr}description
{tabStr}name
{tabStr}{f"memberOrganizations {CollaborativeMemberOrganization.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.collaborative import Collaborative
from ..data_types.data_store import Service


class collaborative_listCollaborativesServiceSharedWith(Query):

    def __init__(self, client, debug, serviceType, serviceId):
        super().__init__(client, debug,
                         "collaborative_listCollaborativesServiceSharedWith", Collaborative)
        if serviceType is None:
            raise MissingParameterException("Missing serviceType")
        if serviceId is None:
            raise MissingParameterException("Missing serviceId")
        self.serviceType = serviceType
        self.serviceId = serviceId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(serviceType: {self.serviceType}, serviceId: "{self.serviceId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(
            f'Listing the collaboratives service {self.serviceId} is shared with...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.collaboratives = []
            for currentCollaborative in self._result['data'][self.queryName]:
                self.collaboratives.append(Collaborative(
                    currentCollaborative, self._client, self._debug))
            i = 0
            for entry in self.collaboratives:
                print(f'{i}: {entry}')
                i += 1
            return self.collaboratives
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals


class TimeSeriesAssetAttributeData:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""
   timestamp: {self.timestamp}
   description: {self.value}"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.timestamp = dict['timestamp']
        self.value = dict['value']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}timestamp
{tabStr}value
{tabStr[0:-4]}}} """



import requests
from . import globals
import json
import base64
import websocket
from .subscriptionClient import Subscription


class Client:

    def __init__(self, token, stage, debug):
        APPSYNC_API_ENDPOINT_URL_dev = 'https://annsvlcb4vew7msipwjyzzvyhi.appsync-api.us-west-2.amazonaws.com/graphql'
        APPSYNC_API_ENDPOINT_URL_prod = 'https://rrfs7pb7ancybo7bom7uxcsaxq.appsync-api.us-west-2.amazonaws.com/graphql'
        self.token = token
        self._debug = debug
        self._stage = stage
        if self._stage == "PRODUCTION":
            self.APPSYNC_API_ENDPOINT_URL = APPSYNC_API_ENDPOINT_URL_prod
        else:
            self.APPSYNC_API_ENDPOINT_URL = APPSYNC_API_ENDPOINT_URL_dev
        return

    def performQuery(self, queryString):
        if self._debug:
            print(f'DEBUG - {queryString}')
        response = None

        try:
            with requests.Session() as session:
                response = session.request(
                    url=self.APPSYNC_API_ENDPOINT_URL,
                    method='POST',
                    headers={'Accept': 'application/json',
                             'Content-Type': 'application/json',
                             'Authorization': self.token},
                    json={'query': queryString}
                )
        except Exception as e:
            raise e
        return response.json()

    def performMutation(self, mutationString):
        if self._debug:
            print(f'DEBUG - {mutationString}')
        response = None

        try:
            with requests.Session() as session:
                # Now we can simply post the request...
                response = session.request(
                    url=self.APPSYNC_API_ENDPOINT_URL,
                    method='POST',
                    headers={'Accept': 'application/json',
                             'Content-Type': 'application/json',
                             'Authorization': self.token},
                    json={'query': mutationString}
                )
        except Exception as error:
            raise error
        return response.json()

    def getSubscriptionClient(self, subscriptionName, subscriptionQueryString, subscriptionMessageQueue, subscriptionErrorQueue):
        if self._debug:
            print(f'subscribing to {subscriptionQueryString}...')
        subscription = Subscription(
            self._debug, subscriptionName, subscriptionQueryString, self.APPSYNC_API_ENDPOINT_URL, self.token, subscriptionMessageQueue, subscriptionErrorQueue)
        return subscription



from ..exceptions import *
from .. import globals


class EndpointTelemetry:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    #     def __str__(self):
    #         return f"""telemetry:
    #     name: {self.name}
    #     type: {self.schemaType}
    # """

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.attributeType = dict['attributeType']
        self.name = dict['name']
        self.description = dict['description']
        self.schemaType = dict['schemaType']
        self.index = dict['index']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}name
{tabStr}attributeType
{tabStr}description
{tabStr}schemaType
{tabStr}index
{tabStr[0:-4]}}} """



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.collaborative import Collaborative


class collaborative_getCollaborative(Query):

    def __init__(self, client, debug, collaborativeId):
        super().__init__(client, debug, "collaborative_getCollaborative", Collaborative)
        if collaborativeId is None:
            raise MissingParameterException("Missing collaborativeId")
        self.collaborativeId = collaborativeId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(collaborativeId: "{self.collaborativeId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(f'Retrieving collaborative {self.collaborativeId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> Collaborative:
        super()._processResult()
        try:
            result = Collaborative(
                self._result['data'][self.queryName], self._client, self._debug)
            print(result)
            return result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.time_series_generate_query_export_result import GenerateQueryResultExportFileResult


class timeSeriesDb_generateQueryResultExportFile(Mutation):

    def __init__(self, client, debug, timeSeriesDbId, queryId, fileFormat):
        super().__init__(client, debug, "timeSeriesDb_generateQueryResultExportFile",
                         GenerateQueryResultExportFileResult)
        self.timeSeriesDbId = timeSeriesDbId
        self.queryId = queryId
        self.fileFormat = fileFormat

    def _getMutationString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        return f"""mutation {self.mutationName} {{ 
{tabStr}{self.mutationName}(
{tabStr}    queryId: "{self.queryId}", 
{tabStr}    fileFormat: {self.fileFormat}, 
{tabStr}    timeSeriesDbId: "{self.timeSeriesDbId}") {self.resultType.getQueryString(tabs+1)} 
}}"""

    def performMutation(self):
        print(f"Requesting query result export...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.response = GenerateQueryResultExportFileResult(
                self._result['data'][self.mutationName], self._client, self._debug)
            print("Server is compiling result export...")
            return self.response
        except Exception as e:
            raise e



from ..exceptions import *
from .. import globals
import requests
from ..data_types.user import User
from ..data_types.data_store_object import DataStoreObject
from ..data_types.data_store_file import DataStoreFile


class DataStoreDirectory(DataStoreObject):
    def __init__(self, dict, client, debug, parentDirectory=None, retrieveContentObjects=True):
        super().__init__(dict, client, debug)
        self.dataStoreDirectory = self
        self._client = client
        self._debug = debug
        self.parentDirectory = parentDirectory
        if dict is None:
            return
        else:
            self.buildFromDictionary(dict, retrieveContentObjects)

    def __str__(self):
        return f"""name: ${self.name}
    dataStoreDirectoryId: {self.dataStoreDirectoryId}
    dataStoreId: {self.dataStoreId}
    parentDirectory: {self.parentDirectoryId}
    contents: 
{"        empty" if len(self.directories) == 0 and len(self.files) == 0 else self.printContents()} """

    def printContents(self):
        result = ""
        if len(self.directories) > 0:
            for dir in self.directories:
                result += f"        {dir.name}/ | id: {dir.dataStoreDirectoryId}\n"
        if len(self.files) > 0:
            for file in self.files:
                result += f"        {file.getFilename()} | id: {file.dataStoreFileId}\n"
        return result

    def buildFromDictionary(self, dict, retrieveContentObjects=True):
        if dict is None:
            raise MissingDataInResultException
        self.dataStoreDirectoryId = dict['dataStoreDirectoryId']
        self.dataStoreId = dict['dataStoreId']
        self.name = dict['name']
        self.ownerUserId = dict['ownerUserId']
        self.homeDirectory = dict['homeDirectory']
        self.parentDirectoryId = dict['parentDirectoryId']
        self.directories = []
        self.files = []
        for currentObject in dict['contents']:
            if currentObject["objectType"] == "FILE":
                file = self._getFileObject(
                    currentObject["dataStoreFile"]["dataStoreFileId"])
                file.directory = self
                self.files.append(file)
            elif currentObject["objectType"] == "DIRECTORY":
                dir = self.getDirectory(
                    self._client, self._debug, currentObject["dataStoreDirectory"]["dataStoreDirectoryId"])
                dir.parentDirectory = self
                self.directories.append(dir)

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}dataStoreDirectoryId
{tabStr}dataStoreId
{tabStr}name
{tabStr}ownerUserId
{tabStr}homeDirectory
{tabStr}parentDirectoryId
{tabStr}{f"contents {DataStoreObject.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """

    def ls(self, path):
        dir = self.cd(path)
        return self.getLsString()

    def getLsString(self):
        result = ".\n"
        if self.homeDirectory is not True:
            result += "..\n"
        for dir in self.directories:
            result += f"{dir.name}/ | id: {dir.dataStoreDirectoryId}\n"
        for file in self.files:
            result += f"{file.getFilename()} | id: {file.dataStoreFileId}\n"
        return result

    def getPath(self):
        if self.homeDirectory:
            self.path = "/"
        else:
            self.path = self.parentDirectory.getPath() + f'{self.name}/'
        return self.path

    def cd(self, path):
        path = path.strip()
        nextDir = None
        if path == "":
            return self
        elif path == "../" or path == "..":
            if self.homeDirectory:
                raise Exception("Invalid path")
                return self
            else:
                return self.parentDirectory
        elif path == "./" or path == ".":
            return self
        elif path.startswith("/"):
            raise Exception("Invalid path")
            return self
        elif path.startswith("../"):
            return self.parentDirectory.cd(path[3:])
        elif path.startswith("./"):
            return self.cd(path[2:])
        else:
            nextDirName = ""
            try:
                names = path.split("/")
                nextDirName = names[0]
            except:
                raise Exception("Invalid path")
                return self
            else:
                for dir in self.directories:
                    if dir.name == nextDirName:
                        nextDir = dir

        if nextDir is not None:
            if nextDir == self:
                return self

            remainingPath = ""
            try:
                names = path.split("/")
                remainingPath = ("/").join(names[1:])
            except Exception as e:
                print(e)
                return nextDir
            else:
                return nextDir.cd(remainingPath)

        raise Exception("Invalid path")
        return self

    def _getFileObject(self, dataStoreFileId):
        from ..graphql_queries.storage_getDataStoreFile import storage_getDataStoreFile
        query = storage_getDataStoreFile(
            self._client, self._debug, dataStoreFileId)
        dataStoreFile = query.performQuery()
        return dataStoreFile

    def downloadFile(self, filename, writeLocation):
        for file in self.files:
            if file.getFilename() == filename:
                url = file.getFileUrl(file.dataStoreFileId)
                location = file.downloadAndWriteFile(url.url, writeLocation)
                return file

        raise Exception("No file matching filename in current directory")

    def deleteFile(self, filename):
        for file in self.files:
            if file.getFilename() == filename:
                file.deleteDataStoreFile(
                    self._client, self._debug, file.dataStoreFileId)
                return file

        raise Exception("No file matching filename in current directory")

    @staticmethod
    def getDirectory(client, debug, dataStoreDirectoryId):
        from ..graphql_queries.storage_getDataStoreDirectory import storage_getDataStoreDirectory
        query = storage_getDataStoreDirectory(
            client, debug, dataStoreDirectoryId)
        result = query.performQuery()
        return result



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.data_store import DataStore
from ..data_types.time_series_db import TimeSeriesDb
from ..data_types.service import Service


class collaborative_addServiceToCollaborative(Mutation):

    def __init__(self, client, debug, collaborativeId, serviceType, serviceId):
        super().__init__(client, debug, "collaborative_addServiceToCollaborative", Service)
        self.collaborativeId = collaborativeId
        self.serviceType = serviceType
        self.serviceId = serviceId

    def _getMutationString(self):
        return f"""mutation {self.mutationName} {{ {self.mutationName}(collaborativeId: "{self.collaborativeId}", serviceType: {self.serviceType}, serviceId: "{self.serviceId}") {self.resultType.getQueryString()} }}"""

    def performMutation(self):
        print("Adding your service to the collaborative...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.service = Service(
                self._result['data'][self.mutationName], self._client, self._debug)
            print("Service was successfully shared to the collaborative")
            return self.service
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of mutation {self.mutationName}")



from ..exceptions import *
from .. import globals
from .service import Service
import requests


class DataStore(Service):
    def __init__(self, dict, client, debug):
        super().__init__(dict, client, debug)
        self.dataStore = self
        self._client = client
        self._debug = debug
        if dict is None:
            return
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""name: {self.name}
   dataStoreId: {self.dataStoreId}
   description: {self.description}
   homeDirectoryId: {self.homeDirectoryId}"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.dataStoreId = dict['dataStoreId']
        self.name = dict['name']
        self.description = dict['description'] if 'description' in dict else ""
        self.ownerUserId = dict['ownerUserId']
        self.homeDirectoryId = dict['homeDirectoryId']

        self.serviceType = "DATA_STORE"
        self.serviceId = self.dataStoreId
        self.currentDirectory = self._getDirectoryObject(
            self.homeDirectoryId)

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}dataStoreId
{tabStr}name
{tabStr}description
{tabStr}ownerUserId
{tabStr}homeDirectoryId
{tabStr[0:-4]}}} """

    def getFileUrl(self, dataStoreFileId):
        from ..graphql_queries.storage_getDataStoreFileDownloadUrl import storage_getDataStoreFileDownloadUrl
        query = storage_getDataStoreFileDownloadUrl(
            self._client, self.dataStoreId, dataStoreFileId)
        result = query.performQuery()
        return result

    @staticmethod
    def downloadAndWriteFile(url, writeLocation):
        response = requests.get(url)
        open(writeLocation, "wb").write(response.content)

    def _getDirectoryObject(self, dataStoreDirectoryId):
        from .data_store_directory import DataStoreDirectory
        return DataStoreDirectory.getDirectory(self._client, self._debug, dataStoreDirectoryId)

    def cd(self, path):
        if path is None:
            raise Exception('path is a required parameter')
        print(
            f'{self.getWorkingDirectoryPath()}: cd {path if path is not None else ""}')
        self.currentDirectory = self.currentDirectory.cd(path)

    def ls(self, path=None):
        lsResult = None
        if path is None:
            lsResult = self.currentDirectory.getLsString()
        else:
            lsResult = self.currentDirectory.ls(path)

        print(f'{self.getWorkingDirectoryPath()}: ls {path if path is not None else ""}\n{lsResult if lsResult is not None else ""}')

    def pwd(self):
        print(
            f'{self.getWorkingDirectoryPath()}: pwd\n{self.getWorkingDirectoryPath()}')

    def getWorkingDirectoryPath(self):
        return self.currentDirectory.getPath()

    @staticmethod
    def help():
        print("Available methods:")
        print(" - ls(path)                    | Print contents of a directory.")
        print("                               | Parameters:")
        print("                               |  - path: string, optional")
        print("                               |      The relative path to the directory to print contents of")
        print("")
        print(" - cd(path)                    | Change working directory")
        print("                               | Parameters:")
        print("                               |  - path: string, required")
        print("                               |      The relative path of the desired directory")
        print("")
        print(" - pwd()                       | Print the path of the working directory.")
        print("")
        print(
            " - uploadFile(localFilePath,   | Upload a file to the data store in the working directory.")
        print("            name,              | Parameters:")
        print("            description)       |  - localFilePath: string, required")
        print("                               |      The location of the file on your local file system.")
        print("                               |  - name: string, optional")
        print("                               |      What the file should be named in the data store.")
        print("                               |  - description: string, optional")
        print("                               |      A brief description of the file and it's contents.")
        print(
            "                               |      Will use local filename if not provided.")
        print("                               |  - description: string, optional")
        print("                               |      A description of what the file is and the data")
        print("                               |      it contains")
        print("")
        print(" - downloadFile(filename,      | Download the specified file.")
        print("                writeLocation) | Parameters:")
        print("                               |  - filename: string, required")
        print("                               |      Name of the remote file in the current directory.")
        print("                               |  - writeLocation: string, required")
        print("                               |      The location to store the downloaded file. If a filename")
        print("                               |      is not given it will be saved with it's remote name.")
        print("")
        print(
            " - deleteFile(filename)        | Deletes the file from the working directory.")
        print("                               | Parameters:")
        print("                               |  - filename: string, required")
        print("                               |      Name of the file to delete.")

    def downloadFile(self, filename, writeLocation):
        return self.currentDirectory.downloadFile(filename, writeLocation)

    def uploadFile(self, localFilePath, name=None, description=None):
        from .data_store_file import DataStoreFile
        file = DataStoreFile.createAndUploadFile(
            self._client, self._debug, localFilePath, self.dataStoreId, self.currentDirectory.dataStoreDirectoryId, name, description)
        self.updateCurrentDirectoryContents()
        return file

    def deleteFile(self, filename):
        file = self.currentDirectory.deleteFile(filename)
        self.updateCurrentDirectoryContents()
        return file

    def updateCurrentDirectoryContents(self):
        parent = self.currentDirectory.parentDirectory
        self.currentDirectory = self._getDirectoryObject(
            self.currentDirectory.dataStoreDirectoryId)
        self.currentDirectory.parentDirectory = parent



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.presigned_url import PresignedUrl
from ..data_types.service import Service


class storage_getDataStoreFileDownloadUrl(Query):

    def __init__(self, client, debug, dataStoreId, dataStoreFileId):
        super().__init__(client, debug, "storage_getDataStoreFileDownloadUrl", PresignedUrl)
        if dataStoreId is None:
            raise MissingParameterException("Missing parameter dataStoreId")
        if dataStoreFileId is None:
            raise MissingParameterException(
                "Missing parameter dataStoreFileId")
        self.dataStoreId = dataStoreId
        self.dataStoreFileId = dataStoreFileId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(dataStoreId: "{self.dataStoreId}", dataStoreFileId: "{self.dataStoreFileId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print('Retrieving file download url...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> PresignedUrl:
        super()._processResult()
        try:
            return PresignedUrl(self._result['data'][self.queryName], self._client, self._debug)
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.iot_query_by_time_range_result import IotQueryByTimeRangeResult
import json


class iot_queryByTimeRange(Query):

    def __init__(self, client, debug, endpointQueryFilters, startTime, endTime, nextToken=None, clientToken=None, queryString=None, queryId=None):
        super().__init__(client, debug, "iot_queryByTimeRange",
                         IotQueryByTimeRangeResult)

        self.endpointQueryFilters = endpointQueryFilters

        self.startTime = f'{startTime}'
        self.endTime = f'{endTime}'

        self.queryString = queryString
        self.nextToken = nextToken
        self.clientToken = clientToken
        self.queryId = queryId

    def getAttributeNamesListString(self, attributeNamesArray):

        result = f'"{attributeNamesArray[0]}"'
        for i in range(1, len(attributeNamesArray)):
            result += f', "{attributeNamesArray[i]}"'
        return result

    def _getQueryString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        endpointFilterQueryString = ""
        for entry in self.endpointQueryFilters:
            endpointFilterQueryString += entry.getMutationParameterString(
                tabs + 1)
        query = f'''query {self.queryName} {{ {self.queryName} ( endpointQueryFilters: [{endpointFilterQueryString}], 
    startTime: "{self.startTime}", 
    endTime: "{self.endTime}"'''
        if self.nextToken:
            query += f', nextToken: "{self.nextToken}"'
        if self.queryString:
            query += f', queryString: {json.dumps(self.queryString)}'
        if self.clientToken:
            query += f', clientToken: "{self.clientToken}"'
        if self.clientToken:
            query += f', queryId: "{self.queryId}"'
        query += f') {self.resultType.getQueryString()} }}'
        return query

    def performQuery(self) -> str:
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> IotQueryByTimeRangeResult:
        super()._processResult()
        try:
            self.result = IotQueryByTimeRangeResult(
                self._result['data'][self.queryName], self._client, self._debug)
            return self.result
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals
from .endpoint_property import EndpointProperty
from .endpoint_telemetry import EndpointTelemetry


class IotEndpoint(object):
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    #     def __str__(self):
    #         return f"""endpoint:
    #     name: {self.name}
    #     iotEndpointId: {self.iotEndpointId}
    #     modelId: {self.modelId}
    # """

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.iotEndpointId = dict["iotEndpointId"]
        self.iotHubId = dict["iotHubId"]
        self.modelId = dict["modelId"]
        self.ownerUserId = dict["ownerUserId"]
        self.name = dict["name"]
        self.description = dict["description"]
        self.properties = []
        for entry in dict["properties"]:
            self.properties.append(EndpointProperty(
                entry, self._client, self._debug))
        self.telemetry = []
        for entry in dict["telemetry"]:
            self.telemetry.append(EndpointTelemetry(
                entry, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}iotEndpointId
{tabStr}iotHubId
{tabStr}modelId
{tabStr}ownerUserId
{tabStr}name
{tabStr}description
{tabStr}{f"properties {EndpointProperty.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}{f"telemetry {EndpointTelemetry.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



from ..exceptions import *
from .. import globals


class TimeSeriesDimension:
    def __init__(self, DimensionValueType, Name, Value):
        self.dimensionValueType = DimensionValueType
        self.name = Name
        self.value = Value
        return

    def __str__(self):
        return f"""      Dimension:
        dimensionValueType: {self.dimensionValueType}
        name: {self.name}
        value: {self.value}
"""

    def getMutationParameterString(self, tabs):
        tabStr = ""
        tab = "    "
        for i in range(tabs):
            tabStr += tab

        return f""" {{
{tabStr}DimensionValueType: {self.dimensionValueType},
{tabStr}Name: "{self.name}",
{tabStr}Value: "{self.value}"
{tabStr[0:-4]}}}"""



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.collaborative import Collaborative
from ..data_types.service import Service


class collaborative_listCollaborativeServices(Query):

    def __init__(self, client, debug, collaborativeId):
        super().__init__(client, debug, "collaborative_listCollaborativeServices", Service)
        self.collaborativeId = collaborativeId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(collaborativeId: "{self.collaborativeId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(
            f'Listing services available in collaborative {self.collaborativeId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            self.services = []
            for currentService in self._result['data'][self.queryName]:
                self.services.append(Service.getCorrectServiceTypeFromServiceObject(
                    currentService, self._client, self._debug))
            i = 0
            for entry in self.services:
                print(f'{i}: {entry}')
                i += 1
            return self.services
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals


class EndpointPropertyInput:
    def __init__(self, name, value, timestamp):
        self.name = name
        self.value = value
        self.timestamp = timestamp
        return

    def getMutationParameterString(self, tabs):
        tabStr = globals.getTabStr(tabs)
        result = f""" {{
{tabStr}name: "{self.name}",
{tabStr}value: "{self.value}",
{tabStr}timestamp: "{self.timestamp}"
{tabStr[0:-4]}}} """
        return result

    @staticmethod
    def checkDictForCorrectFields(dict):
        if "name" not in dict:
            raise InvalidParameterException(
                "Endpoint property entry is missing 'name'")
        if "value" not in dict:
            raise InvalidParameterException(
                "Endpoint property entry is missing 'value'")
        if "timestamp" not in dict:
            raise InvalidParameterException(
                "Endpoint property entry is missing 'timestamp'")
        return True



from .mutation import *
from ..exceptions import *
from .. import globals
from ..data_types.time_series_publish_input import TimeSeriesPublishInput
from ..data_types.time_series_publish_response import TimeSeriesPublishResponse
from ..data_types.time_series_asset_data import TimeSeriesAssetData


class timeSeriesDb_publishToDatabase(Mutation):

    def __init__(self, client, debug, timeSeriesDbId, assetId, records):
        super().__init__(client, debug, "timeSeriesDb_publishToDatabase", TimeSeriesAssetData)
        self.timeSeriesDbId = timeSeriesDbId
        self.assetId = assetId
        self.records = records
        self.dataInput = TimeSeriesPublishInput(self.records)

    def _getMutationString(self):
        tabs = 1
        tabStr = globals.getTabStr(tabs)
        return f"""mutation {self.mutationName} {{ 
{tabStr}{self.mutationName}(
{tabStr}    assetId: "{self.assetId}", 
{tabStr}    timeSeriesDbId: "{self.timeSeriesDbId}", 
{tabStr}    data: {self.dataInput.getMutationParameterString(tabs+2)}) {self.resultType.getQueryString(tabs+1)} 
}}"""

    def performMutation(self):
        print(f"Publishing records to the database {self.timeSeriesDbId}...")
        self._result = self._client.performMutation(self._getMutationString())
        return self._processResult()

    def _processResult(self):
        super()._processResult()
        try:
            # self.response = TimeSeriesPublishResponse(self._result['data'][self.mutationName], self._client)
            self.assetData = []
            for current in self._result['data'][self.mutationName]:
                self.assetData.append(
                    TimeSeriesAssetData(current, self._client, self._debug))
            print("Publish succeeded")
            return self.assetData
        except Exception as e:
            raise e



from ..exceptions import *
from .. import globals
from .iot_data_record import IotDataRecord
from .iot_data_record_with_errors import IotDataRecordWithErrors


class PublishIotDataResult:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    #     def __str__(self):
    #         return f"""Publish Response:
    #     Records Written: {self.recordsWritten}
    # """

    def buildFromDictionary(self, dict):
        self.resultDict = dict
        print(dict)
        if dict is None:
            raise MissingDataInResultException
        self.recordsWritten = []
        for record in dict['recordsWritten']:
            self.recordsWritten.append(
                IotDataRecord(record, self._client, self._debug))
        self.failedRecords = []
        for record in dict['failedRecords']:
            self.failedRecords.append(
                IotDataRecordWithErrors(record, self._client, self._debug))

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}{f"recordsWritten {IotDataRecord.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}{f"failedRecords {IotDataRecordWithErrors.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """



class Unauthorized(Exception):
    def __init__(self, message="Unauthorized to perform the action"):
        self.message = message
        super().__init__(self.message)


class UnauthorizedException(Exception):
    def __init__(self, message="Unauthorized to perform the action. Possibly from an invalid or expired authentication token."):
        self.message = message
        super().__init__(self.message)


class QueryFailed(Exception):
    def __init__(self, message="The query failed.", error=""):
        self.message = f'{message}\nERROR - Query Result/Error: {error}'
        super().__init__(self.message)


class MutationFailed(Exception):
    def __init__(self, message="The mutation failed.", error=""):
        self.message = f'{message}\nERROR - Mutation Result/Error: {error}'
        super().__init__(self.message)


class SubscriptionFailed(Exception):
    def __init__(self, message="The subscription failed.", error=""):
        self.message = f'{message}\nERROR - subscription Result/Error: {error}'
        super().__init__(self.message)


class ParsingResultError(Exception):
    def __init__(self, message="Error encountered while processing the query result."):
        self.message = message
        super().__init__(self.message)


class MissingDataInResultException(Exception):
    def __init__(self, message="Missing expected data in result."):
        self.message = message
        super().__init__(self.message)


class ServiceTypeNotAvailable(Exception):
    def __init__(self, message="This service type is not yet supported in the SDK."):
        self.message = message
        super().__init__(self.message)


class MissingParameterException(Exception):
    def __init__(self, message="Missing required parameter."):
        self.message = message
        super().__init__(self.message)


class FileNotFoundException(Exception):
    def __init__(self, message="File not found in file system"):
        self.message = message
        super().__init__(self.message)


class FileUploadException(Exception):
    def __init__(self, message="File upload failed"):
        self.message = message
        super().__init__(self.message)


class InvalidParameterException(Exception):
    def __init__(self, message="You passed an invalid type to the function."):
        self.message = message
        super().__init__(self.message)


class InvalidTimestampParameter(Exception):
    def __init__(self, parameterName, timestamp):
        self.message = f'Received invalid timestamp for parameter "{parameterName}": "{timestamp}"'
        super().__init__(self.message)




from .data_types.service import Service

from .data_types.collaborative import Collaborative
from .graphql_mutations.collaborative_removeServiceFromCollaborative import collaborative_removeServiceFromCollaborative
from .graphql_mutations.collaborative_addServiceToCollaborative import collaborative_addServiceToCollaborative
from .graphql_queries.collaborative_getCollaborative import collaborative_getCollaborative
from .graphql_queries.collaborative_listCollaborativesServiceSharedWith import collaborative_listCollaborativesServiceSharedWith
from .graphql_queries.collaborative_listCollaborativeServices import collaborative_listCollaborativeServices
from .graphql_queries.collaborative_listCollaboratives import collaborative_listCollaboratives

import time
import signal


class CollaborativeUtil(object):

    def __init__(self, debug, client):
        self._debug = debug
        self._client = client

    def listCollaboratives(self):
        """Lists the Collaboratives the user is a member of"""
        query = collaborative_listCollaboratives(self._client, self._debug)
        result = query.performQuery()
        return result

    def getCollaborative(self, collaborativeId) -> Collaborative:
        """Gets the Collaborative's metadata by collaborativeId"""
        query = collaborative_getCollaborative(
            self._client, self._debug, collaborativeId)
        result = query.performQuery()
        return result

    def listCollaborativeServices(self, collaborativeId):
        """Lists all Services shared in the Collaborative"""
        query = collaborative_listCollaborativeServices(
            self._client, self._debug, collaborativeId)
        result = query.performQuery()
        return result

    def listCollaborativesServiceSharedWith(self, serviceId):
        """Lists the Collaboratives that a Service is shared with"""
        serviceType = 'unknown'
        if 'dataStore' in serviceId:
            serviceType = 'DATA_STORE'
        elif 'timeSeriesDb' in serviceId:
            serviceType = 'TIME_SERIES_DB'
        elif 'iot' in serviceId:
            serviceType = 'IOT_GROUP'
        query = collaborative_listCollaborativesServiceSharedWith(
            self._client, self._debug, serviceType, serviceId)
        result = query.performQuery()
        return result

    def addServiceToCollaborative(self, serviceId, collaborativeId) -> Service:
        """Shares the Service to the Collaborative"""
        serviceType = 'unknown'
        if 'dataStore' in serviceId:
            serviceType = 'DATA_STORE'
        elif 'iotGroup' in serviceId:
            serviceType = 'IOT_GROUP'
        mutation = collaborative_addServiceToCollaborative(
            self._client, self._debug, collaborativeId, serviceType, serviceId)
        result = mutation.performMutation()
        return result

    def removeServiceFromCollaborative(self, serviceId, collaborativeId) -> Service:
        """Removes the Service from the Collaborative"""
        serviceType = 'unknown'
        if 'dataStore' in serviceId:
            serviceType = 'DATA_STORE'
        elif 'iotGroup' in serviceId:
            serviceType = 'IOT_GROUP'
        mutation = collaborative_removeServiceFromCollaborative(
            self._client, self._debug, collaborativeId, serviceType, serviceId)
        result = mutation.performMutation()
        return result



from ..exceptions import *
from .. import globals


class Subscription:
    def __init__(self, client, debug, subscriptionName, resultType):
        self._client = client
        self._debug = debug
        self._result = None
        self.subscriptionName = subscriptionName
        self.resultType = resultType
        return

    def _getSubscriptionQueryString(self):
        return f"subscription {self.subscriptionName} {{ {self.subscriptionName} {self.resultType.getQueryString()} }}"

    def _processResult(self):
        if self._debug:
            print(
                f'DEBUG - processing subscription result: {self.subscriptionName}')
            print(f'DEBUG - {self._result}')
        if 'errors' in self._result and len(self._result['errors']) > 0:
            print('ERROR - Subscription encountered error.')
            print(f'ERROR - Response/Error: {self._result}')
            if 'errorType' in self._result['errors'][0] and self._result['errors'][0]['errorType'] == "UnauthorizedException":
                raise UnauthorizedException
            elif 'message' in self._result['errors'][0]['message'] and self._result['errors'][0]['message'] == "Unauthorized":
                raise Unauthorized
            else:
                raise SubscriptionFailed(
                    f"ERROR - Subscription {self.subscriptionName} failed.", self._result)

        if self._result['data'][self.subscriptionName] is None:
            raise MissingDataInResultException



import os
from .exceptions import *
from .iotUtil import *
from .dataStoreUtil import *
from .collaborativeUtil import *

from .data_types.user import User
from .graphql_queries.user_getUserSession import user_getUserSession
from .client import Client
from . import globals


# from .graphql_queries.storage_getDataStoreFileDownloadUrl import storage_getDataStoreFileDownloadUrl

class AvistaDigitalExchange(object):

    def __init__(self, token, debug=False):
        if type(debug) is not bool:
            raise InvalidParameterException(
                "AvistaDigitalExchange debug parameter must be a bool")
        self._stage = "PRODUCTION"
        self._debug = debug
        self._token = token
        self._client = Client(self._token, self._stage, self._debug)
        self.iot = IoTUtil(self._debug, self._client)
        # self.collaboratives = CollaborativeUtil(self._debug, self._client)
        self.dataStores = DataStoreUtil(self._debug, self._client)
        print("SDK initialized!")

    def getUserInfo(self) -> User:
        """Retrieves the user information of the user associated with the authentication token in use."""
        query = user_getUserSession(self._client, self._debug)
        result = query.performQuery()
        return result


def isValidISO8601Timestamp(timestamp):
    import re

    matched = re.match("\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z", timestamp)
    isMatch = bool(matched)

    return isMatch
    # try:
    #     datetime.fromisoformat(timestamp.replace('Z', '+00:00'))
    # except:
    #     return False
    # return True




from ..exceptions import *
from .. import globals
import requests
import os


class IotQueryExport:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionary(dict)

    def __str__(self):
        return f"""Query Export File Info:
    fileFormat: {self.fileFormat}
    fileSizeBytes: {self.fileSizeBytes}
    url: {self.url}
"""

    def buildFromDictionary(self, dict):
        if dict is None:
            raise MissingDataInResultException
        self.queryId = dict['queryId']
        self.fileFormat = dict['fileFormat']
        self.fileSizeBytes = dict['fileSizeBytes']
        self.url = dict['url']

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}queryId
{tabStr}fileFormat
{tabStr}fileSizeBytes
{tabStr}url
{tabStr[0:-4]}}} """

    def getFilename(self):
        if self.fileFormat == "CSV":
            return "result.csv"
        elif self.fileFormat == "JSON":
            return "result.json"
        else:
            return f"result.{self.fileFormat}"

    def downloadAndWriteFile(self, url, writeLocation=None):
        if writeLocation is None:
            writeLocation = f'./'
        response = requests.get(url)
        fullWritePath = self.createWritePath(writeLocation, self.getFilename())
        open(fullWritePath, "wb").write(response.content)
        print('Wrote query export file to ' + fullWritePath)

    @staticmethod
    def createWritePath(writeLocation, cloudFilename):
        # if location is a directory, append the cloudFilename
        if os.path.isdir(writeLocation):
            return os.path.join(writeLocation, cloudFilename)
        # if not, use the writeLocation
        elif os.path.isfile(writeLocation):
            return writeLocation
        else:
            return writeLocation



from ..exceptions import *
from .. import globals


class DataStoreObject:
    def __init__(self, dict, client, debug):
        self._client = client
        self._debug = debug
        if dict is None:
            raise MissingDataInResultException
        else:
            self.buildFromDictionaryRoot(dict)

    def buildFromDictionaryRoot(self, dict):
        # from .data_store_file import DataStoreFile
        # from .data_store_directory import DataStoreDirectory

        if dict is None:
            raise MissingDataInResultException
        if 'objectType' in dict:
            self.objectType = dict['objectType']
            if self.objectType == "FILE":
                # DataStoreFile(dict['dataStoreFile'], self._client, self._debug)
                self.dataStoreFile = dict['dataStoreFile']
            elif self.objectType == "DIRECTORY":
                # DataStoreDirectory(dict['dataStoreDirectory'], self._client, self._debug)
                self.dataStoreDirectory = dict['dataStoreDirectory']
        elif 'dataStoreDirectoryId' in dict:
            self.objectType = "DIRECTORY"
        elif 'dataStoreFileId' in dict:
            self.objectType = "FILE"
        else:
            raise ServiceTypeNotAvailable

    @staticmethod
    def getQueryString(tabs=1, subobjectsRemaining=4):
        from .data_store_file import DataStoreFile
        from .data_store_directory import DataStoreDirectory
        tabStr = globals.getTabStr(tabs)

        return f""" {{
{tabStr}objectType
{tabStr}{f"dataStoreFile {DataStoreFile.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr}{f"dataStoreDirectory {DataStoreDirectory.getQueryString(tabs + 1, subobjectsRemaining - 1)}" if subobjectsRemaining > 0 else ""}
{tabStr[0:-4]}}} """

    # @staticmethod
    # def getCorrectServiceTypeFromServiceObject(dict, client):
    #     if dict["objectType"] == "FILE":
    #         return DataStoreFile(dict['dataStoreFile'], self._client)
    #     elif dict["objectType"] == "DIRECTORY":
    #         return TimeSeriesDb(dict['dataStoreDirectory'], self._client)



from .query import *
from ..exceptions import *
from .. import globals
from ..data_types.data_store import DataStore


class storage_getDataStore(Query):

    def __init__(self, client, debug, dataStoreId):
        super().__init__(client, debug, "storage_getDataStore", DataStore)
        if dataStoreId is None:
            raise MissingParameterException("Missing dataStoreId")
        self.dataStoreId = dataStoreId

    def _getQueryString(self):
        return f'query {self.queryName} {{ {self.queryName}(dataStoreId: "{self.dataStoreId}") {self.resultType.getQueryString()} }}'

    def performQuery(self) -> str:
        print(f'Retrieving data store {self.dataStoreId}...')
        self._result = self._client.performQuery(self._getQueryString())
        return self._processResult()

    def _processResult(self) -> DataStore:
        super()._processResult()
        try:
            self.dataStore = DataStore(
                self._result['data'][self.queryName], self._client, self._debug)
            return self.dataStore
        except Exception as e:
            raise e
            raise Exception(
                f"Error processing result of query {self.queryName}")



from ..exceptions import *
from .. import globals
from ..data_types.time_series_measure_value import TimeSeriesMeasureValue
from ..data_types.time_series_dimension import TimeSeriesDimension


class TimeSeriesInputRecord:
    def __init__(self, Time: str, TimeUnit: str, Version: int, MeasureName: str, MeasureValueType: str, MeasureValue: str = None, MeasureValues=None, Dimensions=None):
        """An object parameter for publishing to a Time Series Database.

        The parameters/members are capitalized to better match the Timestream API specification.
        Reference https://docs.aws.amazon.com/timestream/latest/developerguide/API_Record.html

        Parameters
        ----------
        Time : str, required
        TimeUnit : str, required
        Version : int, optional
        MeasureName : str, required
        MeasureValueType : str, required
        MeasureValue : str, optional
        MeasureValues : [TimeSeriesMeasureValue], optional
        Dimensions : [TimeSeriesDimension], optional
        """

        self.time = Time
        self.timeUnit = TimeUnit
        self.version = Version
        self.measureName = MeasureName
        self.measureValueType = MeasureValueType
        self.measureValue = MeasureValue
        self.measureValues = MeasureValues
        if self.measureValues is not None:
            for measure in self.measureValues:
                if not isinstance(measure, TimeSeriesMeasureValue):
                    InvalidParameterException(
                        "Element in measureValues array is an incorrect type.")
        self.dimensions = Dimensions
        if self.dimensions is not None:
            for dimension in self.dimensions:
                if not isinstance(dimension, TimeSeriesDimension):
                    InvalidParameterException(
                        "Element in dimensions array is an incorrect type.")
        else:
            self.dimensions = []
        return

    def __str__(self):
        return f"""Time Series Record:
   time: {self.time}
   timeUnit: {self.timeUnit}
   version: {self.version}
   measureName: {self.measureName}
   measureValueType: {self.measureValueType}
   measureValue: {self.measureValue}
   measureValues: 
""" + self.getMeasureValuesStr() + f"""   dimensions: 
""" + self.getDimensionsStr()

    def getMeasureValuesStr(self):
        result = ""
        for measure in self.measureValues:
            result += f"{measure}"
        return result

    def getDimensionsStr(self):
        result = ""
        for dimension in self.dimensions:
            result += f"{dimension}"
        return result

    def getMutationParameterString(self, tabs):
        tabStr = ""
        tab = "    "
        for i in range(tabs):
            tabStr += tab

        return f""" {{
{f'{tabStr}Dimensions: [{",".join([dimension.getMutationParameterString(tabs+1) for dimension in self.dimensions])}],' if self.dimensions is not None and len(self.dimensions) > 0 else 'Dimensions: [],'}
{tabStr}MeasureName: "{self.measureName}",
{f'{tabStr}MeasureValue: "{self.measureValue}",' if self.measureValueType != "MULTI" else
f'{tabStr}MeasureValues: [{",".join([measureValue.getMutationParameterString(tabs+1) for measureValue in self.measureValues])}],'}
{tabStr}MeasureValueType: {self.measureValueType},
{tabStr}Time: "{self.time}",
{tabStr}TimeUnit: {self.timeUnit},
{f'{tabStr}Version: {self.version}' if self.version is not None else ''}
{tabStr[0:-4]}}}"""



from ..exceptions import *
from .. import globals
from .iot_attribute_value_input import IotAttributeValueInput
import time
from datetime import datetime


class IotDataRecordInput:
    def __init__(self, timestamp, timeUnit, attributes):
        if timeUnit == "ISO8601":
            self.timeUnit = "MILLISECONDS"
            date = datetime.strptime(
                timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')
            self.timestamp = f'{int((date - datetime(1970, 1, 1)).total_seconds()*1000)}'
        else:
            self.timestamp = timestamp
            self.timeUnit = timeUnit

        self.attributes = []
        if attributes:
            self.addAttributes(attributes)
        return

    def addAttributes(self, attributes):
        # Check that records are of the correct type
        for attribute in attributes:
            if not isinstance(attribute, IotAttributeValueInput):
                raise InvalidParameterException(
                    "Attributes array contains elements that are not of type IotAttributeValueInput")
        self.attributes.extend(attributes)

    def getMutationParameterString(self, tabs):
        tabStr = globals.getTabStr(tabs)
        result = f""" {{
{tabStr}timestamp: "{self.timestamp}",
{tabStr}timeUnit: {self.timeUnit},
{tabStr}attributes: [{','.join([attribute.getMutationParameterString(tabs+1) for attribute in self.attributes])}]
{tabStr[0:-4]}}} """
        return result

    @staticmethod
    def checkDictForCorrectFields(obj):
        if "attributes" not in obj:
            raise InvalidParameterException(
                "Record is missing 'data' array")
        elif type(obj["attributes"]) is not dict:
            raise InvalidParameterException(
                "Record 'attributes' should be of type dict."
            )
        return dict
